/***************************** AfhFMcontext.h *******************************/
/*                                                                          */
/*  FM Execution Environment (Context) definitions.                         */
/*                                                                          */
/*  Copyright 1998-2014 AFH Systems                                         */
/*                                                                          */
/*  FilterMeister is free software. You can redistribute and modify it      */
/*  under the terms of the MPL2 License.                                    */
/*  See: https://www.mozilla.org/en-US/MPL/2.0/                             */
/*  Any changes to the FilterMeister files have to be published or sent     */
/*  to harry@thepluginsite.com. The filter plugin code itself does not      */
/*  need to be disclosed.                                                   */
/*                                                                          */
/*    13Jan98  Alex Hunter  1. Initial version.                             */
/*	  02AUG13  Alex Hunter  1. Changes for WIN64.							*/
/*	  03AUG13  Alex Hunter  1. Use FMint/FMuint to postpone decisions for   */
/*							   the FM ABI prototypes.						*/
/*							2. Define TARGET=32 for x86 targets, or			*/
/*							   TARGET=64 for x64 targets; if TARGET is not  */
/*							   defined at entry, it will be set to 64 if    */
/*							   _WIN64 is defined else it will be set to 32. */
/*	  16AUG13  Alex Hunter  1. Add _ftol2, _ftol2_sse.						*/
/*	  23AUG13  Alex Hunter  1. Add quickFillArray.							*/
/*	  02JAN14  Alex Hunter  1. Add BigDocument 32-bit fields to FMcontext.	*/
/*	  10MAR14  Alex Hunter  1. Add members to FMcontext: abort_mode and		*/
/*							   FMC_compile_platform.						*/
/*							2. Change FMC_MAJOR_VERSION to 2.				*/
/*	  08MAY14  Alex Hunter	1. Rename fmfd to fmfdi.						*/
/*							2. Add fmf and fmfd function types.				*/
/*							3. Add fmf2I and fmfiIi function types.			*/
/*							4. Add comments explaining function type names.	*/
/*	  24MAY14  Alex Hunter	1. Define wf0U in place of wf0.					*/
/*                                                                          */
/*--------------------------------------------------------------------------*/

// Override code generation target if TARGET is set, else set
// TARGET to 32 or 64, depending on whether _WIN64 is defined
#ifndef TARGET
#ifdef _WIN64
#define TARGET 64
#else
#define TARGET 32
#endif
#endif //TARGET

/******************* Context Structure Definition ************************/

// Macros to define types which change depending on the code generation
// target rather than the host.

#undef FMint		//in case previously defined
#undef FMuint		//in case previously defined
#undef FMcontext	//in case previously defined
#undef XINT_PTR		//int size of ptr on target
#undef XUINT_PTR	//uint size of ptr on target

#if TARGET == 64
#define FMint INT64
#define FMuint UINT64
#define XINT_PTR INT64
#define XUINT_PTR UINT64
#else //TARGET == 32
#define FMint INT32
#define FMuint UINT32
#define XINT_PTR INT32
#define XUINT_PTR UINT32
#endif

// Default FMcontext depends on host, not on target...
#ifdef _WIN64
#define FMcontext FMcontext64
#else //Win32
#define FMcontext FMcontext32
#endif

// Chameleon types for FM context record fields.
// When cross-compiling, these are defined to be an integer
// the size of an INT_PTR in the TARGET environment
// (aka an XINT_PTR, or cross-compilation integer the
// size of a pointer).

// When not cross-compiling, these are the names of the
// function types derived from the prototypes of the
// various FM (or Windows) API functions. These names
// are fairly cryptic and not very consistent, since
// the naming conventions grew over time in a haphazard
// manner. But here's a quick cheatsheet for the interpretation
// of the function type names:
// The first two or three letters define the api calling type:
// fmf* - is an "FM api Function" (cdecl for x86)
// wf*	- is a "Windows api Function" (stdcall for x86)
// The following letters designate the parameter types from
// left to right:
//	I	- is an "int" parameter
//	i	- is an "INT_PTR" parameter
//	U	- is an "unsigned int" parameter
//	u	- is a "UINT_PTR" parameter	
//	d	- is a "double" parameter
//	h	- is a "HANDLE" parameter
//	c	- is a "CTLACTION" parameter
// N.B. The final letter is sometimes the type of the function result,
//		not the type of the last parameter.  This can usually be
//		disambiguated from context, or sometimes 'f' is used to
//		explicitly designate a "double" (float) return type.
//		The letter 'n' specifies a void ("null") return type.
//		If no result type is specified, INT_PTR is assumed.
// A 'v' following the list of parameter types means the function
// is variadic, and the position of the 'v' represents the ellipsis
// (...) following the last parameter.
// A decimal number represents the number of repetitions of the
// following parameter designator. If the parameter designator
// is missing, 'i' is assumed. Thus '3i' (or just '3') denotes
// three INT_PTR parameters, and '2d' (or 'dd') represents two
// double parameters. 'fmf0' means an FM function ('fmf') with
// no parameters ('0'), whereas 'fmf' means an FM function with
// unspecified (unprototyped) parameters. In both cases, the return
// type is assumed to be 'INT_PTR'.
// Note that there are many exceptions to these rules, since the
// names were often created before the rules were defined.
// For example, fmf3f designates the type of an FM function with
// two INT_PTR and one double parameter, and result type INT_PTR;
// not a function with three INT_PTR parameters as one would
// expect from strict adherence to the rules. The important aspect
// is that each function type designator should be associated with
// a unique function signature. The interpretation rules given above
// are merely helpful hints.


// First we undef any previous macro definition...
#undef FMVOIDP		// chameleon "void *" type
#undef FMBYTEP		// chameleon "BYTE *" type
#undef FMBufferID	// chameleon "BufferID" type
#undef FMHWND		// chameleon "HWND" type
#undef fmf			// unprototypes args returning INT_PTR
#undef fmfd			// unprototypes args returning double
#undef fmf0
#undef fmf1
#undef fmf2
#undef fmf3
#undef fmf4
#undef fmf5
#undef fmf6
#undef fmf7
#undef fmf8
#undef fmf9
#undef fmf10
#undef fmf11
#undef fmf12
#undef fmf13
#undef fmf14
#undef fmf15
#if 0	//unused
#undef fmf16
#undef fmf17
#undef fmf18
#undef fmf19
#undef fmf20
#endif	//unused
#undef fmf1vd
#undef fmf1v
#undef fmf2v
#undef fmf3v
#undef fmf8v
#undef fmf2I
#undef fmfiIi
#undef fmfdi
#undef fmf2di
#undef fmf1d
#undef fmf2d
#undef fmf4f
#undef fmf1dd
#undef fmf2dd
#undef fmf2ddv
#undef fmfdid
#undef fmfid
#undef fmf0n
#undef fmf1n
#undef fmf1un
#undef fmf5d
#undef fmf4d
#undef fmf3f
#undef fmf5f
#undef fmf10d
#undef fmf1u
#undef fmf2u
#undef fmf3u
#undef fmf4u
#undef fmf5u
#undef fmf6u
#undef fmf5u3
#undef fmf5u3v
#undef fmf7iui
#undef fmf1ca
#undef fmf3uf
#undef fmf6iud
#undef fmf5d2
#undef wf0U
#undef wf0h
#undef wf4

#if TARGET == 64 && defined(_WIN64) || TARGET == 32 && !defined(_WIN64)
// not cross-compiling...
typedef void *FMVOIDP;
typedef BYTE *FMBYTEP;
typedef BufferID FMBufferID;
typedef HWND FMHWND;
typedef FMint (*fmf)();
typedef double (*fmfd)();
typedef FMint (*fmf0)(void);
typedef FMint (*fmf1)(FMint);
typedef FMint (*fmf2)(FMint,FMint);
typedef FMint (*fmf3)(FMint,FMint,FMint);
typedef FMint (*fmf4)(FMint,FMint,FMint,FMint);
typedef FMint (*fmf5)(FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf6)(FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf7)(FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf8)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf9)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf10)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf11)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf12)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf13)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf14)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf15)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
#if 0	//unused
typedef FMint (*fmf16)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf17)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf18)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf19)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
typedef FMint (*fmf20)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint);
#endif	//unused
typedef double (*fmf1vd)(FMint,...);
typedef FMint (*fmf1v)(FMint,...);
typedef FMint (*fmf2v)(FMint,FMint,...);
typedef FMint (*fmf3v)(FMint,FMint,FMint,...);
typedef FMint (*fmf8v)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,FMint,...);
typedef FMint (*fmf2I)(int,int);
typedef FMint (*fmfiIi)(FMint,int,FMint);
typedef FMint (*fmfdi)(double);
typedef FMint (*fmf2di)(double,double);
typedef double (*fmf1d)(FMint);
typedef double (*fmf2d)(FMint,FMint);
typedef double (*fmf4f)(FMint,FMint,FMint,FMint);
typedef double (*fmf1dd)(double);
typedef double (*fmf2dd)(double,double);
typedef double (*fmf2ddv)(double,double,...);
typedef double (*fmfdid)(double,FMint);
typedef FMint (*fmfid)(FMint,double);
typedef void (*fmf0n)(void);
typedef void (*fmf1n)(FMint);
typedef void (*fmf1un) (FMuint);
typedef FMint (*fmf5d)(double,double,FMint,FMint,FMint);
typedef FMint (*fmf4d)(FMint,double,double,FMint);
typedef FMint (*fmf3f)(FMint,FMint,double);
typedef FMint (*fmf5f)(FMint,FMint,FMint,FMint,double);
//Added by Harald Heim for threeDto2D()
typedef FMint (*fmf10d)(FMint,FMint,FMint,FMint,FMint,FMint,FMint,double,double,double);
typedef FMint (*fmf1u)(FMuint);
typedef FMint (*fmf2u)(FMint,FMuint);
typedef FMint (*fmf3u)(FMint,FMint,FMuint);
typedef FMint (*fmf4u)(FMint,FMint,FMint,FMuint);
typedef FMint (*fmf5u)(FMint,FMint,FMint,FMint,FMuint);
typedef FMint (*fmf6u)(FMint,FMint,FMint,FMint,FMint,FMuint);
typedef FMint (*fmf5u3)(FMint,FMint,FMuint,FMint,FMint);
typedef FMint (*fmf5u3v)(FMint,FMint,FMuint,FMint,FMint,...);
typedef FMint (*fmf7iui)(FMint,FMint,FMint,FMint,FMuint,FMuint,FMint);
typedef FMint (*fmf1ca)(CTLACTION);
typedef FMint (*fmf3uf)(FMuint,FMuint,double);
typedef FMint (*fmf6iud)(FMint,FMint,FMuint,double,double,FMuint);
typedef FMint (*fmf5d2)(FMint,FMint,FMint,FMint,double);
//Win32 API uses __stdcall
typedef DWORD  (WINAPI *wf0U)(void);  // wf0U is call-compatible with fmf0, but tool-chain doesn't know this!!
typedef HANDLE (WINAPI *wf0h)(void);  // ditto wf0h
typedef FMint  (WINAPI *wf4)(FMint,FMint,FMint,FMint);

#else
// cross compiling
#define FMVOIDP XINT_PTR
#define FMBYTEP XINT_PTR
#define FMBufferID XINT_PTR
#define FMHWND XINT_PTR
#define fmf XINT_PTR
#define fmfd XINT_PTR
#define fmf0 XINT_PTR
#define fmf1 XINT_PTR
#define fmf2 XINT_PTR
#define fmf3 XINT_PTR
#define fmf4 XINT_PTR
#define fmf5 XINT_PTR
#define fmf6 XINT_PTR
#define fmf7 XINT_PTR
#define fmf8 XINT_PTR
#define fmf9 XINT_PTR
#define fmf10 XINT_PTR
#define fmf11 XINT_PTR
#define fmf12 XINT_PTR
#define fmf13 XINT_PTR
#define fmf14 XINT_PTR
#define fmf15 XINT_PTR
#if 0	//unused
#define fmf16 XINT_PTR
#define fmf17 XINT_PTR
#define fmf18 XINT_PTR
#define fmf19 XINT_PTR
#define fmf20 XINT_PTR
#endif	//unused
#define fmf1vd XINT_PTR
#define fmf1v XINT_PTR
#define fmf2v XINT_PTR
#define fmf3v XINT_PTR
#define fmf8v XINT_PTR
#define fmf2I XINT_PTR
#define fmfiIi XINT_PTR
#define fmfdi XINT_PTR
#define fmf2di XINT_PTR
#define fmf1d XINT_PTR
#define fmf2d XINT_PTR
#define fmf4f XINT_PTR
#define fmf1dd XINT_PTR
#define fmf2dd XINT_PTR
#define fmf2ddv XINT_PTR
#define fmfdid XINT_PTR
#define fmfid XINT_PTR
#define fmf0n XINT_PTR
#define fmf1n XINT_PTR
#define fmf1un XINT_PTR
#define fmf5d XINT_PTR
#define fmf4d XINT_PTR
#define fmf3f XINT_PTR
#define fmf5f XINT_PTR
#define fmf10d XINT_PTR
#define fmf1u XINT_PTR
#define fmf2u XINT_PTR
#define fmf3u XINT_PTR
#define fmf4u XINT_PTR
#define fmf5u XINT_PTR
#define fmf6u XINT_PTR
#define fmf5u3 XINT_PTR
#define fmf5u3v XINT_PTR
#define fmf7iui XINT_PTR
#define fmf1ca XINT_PTR
#define fmf3uf XINT_PTR
#define fmf6iud XINT_PTR
#define fmf5d2 XINT_PTR
#define wf0U XINT_PTR
#define wf0h XINT_PTR
#define wf4 XINT_PTR
#endif


#define FMC_MAJOR_VERSION 2
#define FMC_MINOR_VERSION 0

typedef struct ///FMcontext_tag
{
    /////// FROZEN SECTION STARTS HERE ///////
    /* Undefined variables default to this field: */
    double undefined;
    /*const*/ int FMC_minor_version;
    /*const*/ int FMC_major_version;
    /*const*/ int FMC_struct_size;
	/*const*/ int FMC_target;			// 32 or 64
	/*const*/ int FMC_compile_platform;	// 32 or 64
    /////// FROZEN SECTION ENDS HERE ///////
    /* Temporary variable(s) */
    double temp;
    /* Public FF built-in variables */
    int r, g, b, a; //MUST be contiguous!
    int R, G, B, A; //MUST be contiguous!
    int c;
    int C;
    int i, u, v;
    int I, U, V;
    int x, y, z;
    int X, Y, Z;
    int d, m;
    int D, M;
    /* Public FF built-in variables for Premiere compatibility */
    int r0, g0, b0, a0; //same as r, g, b, a.
    int r1, g1, b1, a1; //same as r, g, b, a.
    int c0, c1;         //same as c.
    int /*i0,*/ u0, v0;     //same as i, u, v (i0 is in globvars)
    int /*i1,*/ u1, v1;     //same as i, u, v (i1 is in globvars)
    int d0, m0;         //same as d, m.
    int d1, m1;         //same as d, m.
    int rmax, gmax, bmax, amax; //same as R, G, B, A.
    int rmin, gmin, bmin, amin; //always 0.
    int cmax, cmin;             //same as C, 0.
    int imax, umax, vmax;       //same as I, U, V.
    int imin, umin, vmin;       //always 0.
    int xmax, ymax, zmax;       //same as X, Y, Z.
    int xmin, ymin, zmin;       //always 0.
    int dmax, mmax;             //same as D, M.
    int dmin, mmin;             //always 0.
    int p, pmax, pmin;  //same as z, zmax, zmin.
    int t, tmin, tmax, total;   // 0, 0, 1, 1.
    /* Public FF+ built-in variables */
    int n;      // control index in Ctl[n] handler
    int e;      // FM event code
    
    //Added by Harald Heim, May 22. 2003
    int previous;      // FM event value			//FMint ???

    int ctlMsg;
    int ctlCode;
    CTLCLASS ctlClass;
    FMint dividend;	//X64 INT_PTR
    FMint divisor;	//X64 INT_PTR
    int scaleFactor;
	int enlargeFactor;
	int scrollFactor;
    int row;
    int rows;
    int col;
    int columns;
    int plane;
    int inPlanes;
    int srcRowDelta;
    int dstRowDelta;
    int mskRowDelta;
    int srcColDelta;
    int dstColDelta;
    int mskColDelta;
    int writeAlpha;
    int alphaMask;      // 0 or 0xff
// Following members will be copied to the standalone filter via the FMCODE resource... /*GLOBAL?*/
    int need_iuv;		//must be first to be copied!!
    int need_d;
    int need_m;
    int need_tbuf;
    int need_t2buf;
	int need_t3buf;
	int need_t4buf;
    int needPremiere;   //Need Premiere compatibility support
    int isTileable;     /*GLOBAL?*/
    int needPadding;    /*GLOBAL?*/
    int bandWidth;      /*GLOBAL?*/
    int use_MMX;
	int use_SSE;
    int use_CMOV;		//must be last to be copied!!
// End of range of members that will be copied to standalone filter.
    int doingProxy;
    int doingScripting;
	int abort_mode;
    int x_start, x_end;
    int y_start, y_end;
    FMBYTEP tbuf;         // (Temporary) tile buffer.    /*GLOBAL?*/	//X64
    FMBYTEP t2buf;        // (Temporary) tile buffer 2.
	FMBYTEP t3buf;        // (Temporary) tile buffer 3.
	FMBYTEP t4buf;        // (Temporary) tile buffer 4.
	FMBufferID tbufID;                                            /*GLOBAL?*/	//X64
	FMBufferID t2bufID;
	FMBufferID t3bufID;
	FMBufferID t4bufID;

	int DESIGNTIME;

	int entryPoint;		//what is this used for??

    /* Some handy FF+ predeclared variables, preserved across invocations */
    /* Also contains user-declared global and static variables.           */
    PREDEF_GLOBALS globvars;                                    /*GLOBAL*/

    /* Private FF+ built-in variables */
    FMHWND hDlg;	//X64
    FMBYTEP srcPtr;	//X64
    FMBYTEP dstPtr;
    FMBYTEP mskPtr;
    FMBYTEP mskData;

	/* BigDocument fields from Host, for 32-bit coordinate support */
	int	bigDocument;

	VPoint	imageSize32;
    VRect	filterRect32;
    int		filterRectWidth32;
    int		filterRectHeight32;
    VRect	inRect32;
    VRect	outRect32;
    VRect	maskRect32;
	VPoint	floatCoord32;
    VPoint	wholeSize32;

    /* Read-only information from host's Filter Record */
    int hostSerialNumber;
    int displayDialog;

    int planes;
    int planesWithoutAlpha;
    int maxSpace;
	int bufferSpace;
    BOOL isFloating;
    BOOL haveMask;

    //Added by Harald Heim, April 09, 2002
    BOOL autoMask;
    
    COLORREF backColor;     //background color in image color space
    COLORREF foreColor;     //foreground color in image color space
    COLORREF bgColor;       //background color in RGB color space
    COLORREF fgColor;       //foreground color in RGB color space
    int hostSig;

    //Added by Harald Heim, January 2003
    FMVOIDP platformData;		//X64
    double imageHRes;
    double imageVRes;

    int imageMode;
    int filterCase;
    BOOL samplingSupport;
    int zoomFactor;
    /* Public FF+ built-in strings and arrays */            /*GLOBAL?*/
    char filterAuthorText[256];
    char filterCategoryText[256];
    char filterTitleText[256];
    char filterFilenameText[256];
    char filterCopyrightText[256];
    char filterDescriptionText[256];
    char filterOrganizationText[256];
    char filterURLText[256];
    char filterVersionText[256];
    char filterCaseText[256];
    char filterImageModeText[256];
    char filterHostText[256];
    char filterInstallDir[256];
    char filterUniqueID[37];				//X64 alignment???

    /* Double-precision floating-point literal pool. */     /*GLOBAL?*/
    /*const*/ int ndlits;
    /*const*/ double dlitPool[MAX_DLITS];                   /*GLOBAL because of size?and can interweave w/user globals*/
    /* Some handy floating-point constants */
    /*const*/ double magic5;    // = 0.49999999999999 (carefully tuned for in-line OP_SQR!)
    /*const*/ double d512byPi;  // = 512.0/Pi (for c2d etc.)
    /*const*/ float  float1p0;  // = 1.0F (single-precision 32-bit float, for in-line powi etc.)
    /* NDP control words */
    int current_cw;
    /*const*/ int CW_FLOOR;
    /*const*/ int CW_CEIL;
    /*const*/ int CW_ROUND;
    /*const*/ int CW_CHOP;
    /* Vectors to public FF built-in functions */	//X64
    fmf3 src;
    fmf3 rad;
    fmf1 ctl;
    fmf3 val;
    fmf2 map;
    fmf2 min;
    fmf2 max;
    fmf2ddv fmin;
    fmf2ddv fmax;
    fmf1 abs;
    fmf3 add;
    fmf3 sub;
    fmf2 dif;
    fmf2 rnd;
    fmf1 rst;
    fmf4 mix;
    fmf4 mix1;
    fmf4 mix2;
    fmf5 scl;
    fmf1 sqr;
    fmf1 sin;
    fmf1 cos;
    fmf1 tan;
    fmf2 r2x;
    fmf2 r2y;
    fmf2 c2d;
    fmf2 c2m;
    fmf1 get;
    fmf2 put;
    fmf10 cnv;
    /* Vectors to public FF+ built-in functions */

    //added by Harald Heim, February 4, 2002
    fmf4 pointer_to_buffer;
    //added by Harald Heim, February 5, 2002
    fmf2 contrast;
    fmf5 saturation;
    fmf2 posterize;
    fmf2 phaseshift;
    fmf5 blend;
    //added by Harald Heim, February 6, 2002
    fmf4 rgb2iuv;
    fmf3 egw;
    fmf6 gray;
    fmf2 solarize;
    fmf2 msk;
    //added by Harald Heim, February 7, 2002
    fmf1 tricos;
    fmf1 tri;
    fmf7 grad2D;
    //added by Harald Heim, February 8, 2002
    fmf1 sinbell;
    fmf4 iuv2rgb;
    fmf4 tone;
    fmf4 rgb2cmyk;
    fmf5 cmyk2rgb;
    //added by Harald Heim, February 10, 2002
    fmf13 xyzcnv;
    fmf1 cell_initialize;
    //added by Harald Heim, February 13, 2002
    fmf1 set_edge_mode;
    fmf1 set_bitdepth_mode;
    fmf1 set_psetp_mode;
	fmf1 set_array_mode;
    fmf5 bCircle;
    fmf6 bRect;
    fmf6 bRect2;
    fmf5 bTriangle;
    //added by Harald Heim, March 7, 2002
    fmf5d iget;
    //added by Harald Heim, March 10, 2002
    fmf0 refreshWindow;
    //Added by Harald Heim, April 02, 2002
    fmf1 setZoom;
	fmf2 setZoomEx;
    //Added by Harald Heim, April 06, 2002
    fmf1 lockWindow;
    fmf1 refreshCtl;
    //fmf2 lockCtl;
    fmf1 refreshRgn;
    //Added by Harald Heim, April 09, 2002
    fmf5 allocArray;
    fmf6 allocArrayPad;
    fmf1 freeArray;
    fmf4 getArray;
    fmf4 UNSAFE_getArray;
	fmf4f fgetArray;
	fmf4f UNSAFE_fgetArray;
    fmf5 putArray;
    fmf5 UNSAFE_putArray;
	fmf5d2 fputArray;
	fmf5d2 UNSAFE_fputArray;

	//fmf5d2 fputArrayLE;
	//fmf4f fgetArrayLE;
	
    
    fmf2 getArrayString;
    fmf3 putArrayString;
    
    //Added by Harald Heim, Jun 4, 2002
    fmf1 calcSBD;
    fmf0 freeSBD;
    fmf2 getSBD;
    fmf3 getSBDangle;
    fmf2 getSBDX;
    fmf2 getSBDY;
    //Added by Harald Heim, Jun 6, 2002
    fmf1 mouseOverWhenInvisible;
    //Added by Harald Heim, Aug 9, 2002
    fmf0 getPreviewCoordX;
    fmf0 getPreviewCoordY;
    //Added by Harald Heim, Aug 12, 2002
    fmf1 getAsyncKeyState;
    fmf1 getAsyncKeyStateF;
    //Added by Harald Heim, Aug 14, 2002
    //fmf10d threeDto2D_X;
    //fmf10d threeDto2D_Y;
    //Added by Harald Heim, Oct 25, 2002
    fmf2 getArrayDim;
    fmf2 copyArray;
    fmf2 fillArray;
    fmfid ffillArray;
    fmf1 ctlEnabled;
    fmf1 ctlEnabledAs;
    //Added by Harald Heim, Dec 7, 2002
    fmf4 rgb2lab;
    fmf4 lab2rgb;
    //Added by Harald Heim, Dec 13, 2002
    fmf1 setClickDrag;
    fmf3 setTimerEvent;
    fmf1 getDisplaySettings;
    //Added by Harald Heim, Dec 14, 2002
    fmf3 scrollPreview;
    //Added by Harald Heim, Dec 23, 2002
    fmf4 rgb2hsl;
    fmf4 hsl2rgb;
    fmf4 ycbcr2rgb;
    fmf4 rgb2ycbcr;
    //Added by Harald Heim, January 2003
    fmf2 srcp;
    fmf3 psetp;
    fmf2 pgetp; 
    fmf0 getWindowsVersion;
    fmf1 setPreviewCursor;
	fmf0 getPreviewCursor;
    fmf3 tsetp;
    fmf2 tgetp;
    //Added by Harald Heim, March 9, 2003
    fmf3 t2setp;
	fmf2 t2getp;
	fmf3 t3setp;
    fmf2 t3getp;
	fmf3 t4setp;
    fmf2 t4getp;
    //Added by Harald Heim, March 12, 2003
    fmf1 deleteRgn;
    //Added by Harald Heim, May 1, 2003
    fmf2 setCtlDivisor;
    //Added by M.W. van der Lee, June 29, 2008
	fmf1 getCtlDivisor;
    fmf2 setCtlGamma;

    fmf2 getCtlCoord;
    fmf2 getCtlRange;
    fmf2 getCtlPos;

	fmf0 countProcessors;



#ifdef THEPLUGINSITE
    fmf9 cnvX_edge;
    fmf11 cnvY_edge;

    fmf4d igetArray;

    fmf1 ct2R;
    fmf1 ct2G;
    fmf1 ct2B;

    fmf3 rgb2ct;

    //fmf4 quickGauss;
    fmf4 fastMotion;
    fmf4 fastBox;
    fmf3 fastGauss;
	fmf3 fastPreciseGauss;
    fmf3 fastMotionGauss;
    //fmf2 getKey;
    fmf1 copyImage2Array;

	fmf7 calcEdgeMask;
	fmf2 stretchContrast;
	fmf2 calcYCC;

    fmf4 floodFill;
    fmf3 noise;

    fmf3 setPixelDB;
    fmf1 startSetPixelDB;
    fmf1 endSetPixelDB;
    fmf3 drawTextDB;
    fmf5 setFontDB;

    fmf10 calcSpecialMinMax;
    fmf8 calcSpecialMinMaxFast;
    fmf10 calcTheMix;
    fmf8 calcCMFinal;
    fmf6 dynaContra;

	fmf6 processHDR;
	fmf5 processHDR2;

	fmf11 processACE;

#endif

    fmf2 quickSort;
    fmf2 quickMedian;
    fmf9 quickFill;
	fmf9 quickFillArray;


#ifdef IMGSOURCE
    fmf1 openImage;
    fmf2 drawArray;
    fmf4 applyFX;
#endif

#if WATERMARK
    fmf2 watermark;
#endif


	fmf3 copyResToArray;
	fmf2 getResAddress;
	fmf2 getResSize;

    fmf0 getDialogWidth;
    fmf0 getDialogHeight;
    fmf1 setDialogShowState;
    fmf1 PixelsToHDBUs;
    fmf1 PixelsToVDBUs;
    fmf4 setDialogMinMax;
    fmf1 setDialogSizeGrip;

    fmf1 startSetPixel;
    fmf1 endSetPixel;
	fmf3u setPixel;
    fmf3 setBitmap;							// mwvdlee 2008-11-12
	fmf4u setBitmapTransparent;				// mwvdlee 2008-11-13
	fmf5 setBitmapStretch;					// mwvdlee 2008-11-13
	fmf6u setBitmapStretchTransparent;		// mwvdlee 2008-11-23
	fmf6 setBitmapTile;						// mwvdlee 2008-11-23
//	fmf6u setBitmapTileTransparent;			// mwvdlee 2008-11-13
//	fmf8 setBitmapTileStretch;				// mwvdlee 2008-11-13
//	fmf8u setBitmapTileStretchTransparent;	// mwvdlee 2008-11-
//	fmf1 setStretchMode;					// mwvdlee 2008-11-13

	fmf2 getPixel;
	fmf5u3 setText;
	fmf5u3v setTextv;
    fmf5 setFont;
	fmf5u setLine;
	fmf5u setEllipse;
	fmf5u setEllipseFill;
	fmf5u setRectFrame;
	fmf5u setRectangle;
	fmf1u setFill;
	fmf5u setRectFill;
	fmf7iui setRectGradient;
	fmf1 setPenWidth;
	fmf2 startSetPixelSS;
	fmf2 startSetPixelBuffer;
	fmf0 endSetPixelBuffer;
	fmf0 getSetPixelWidth;
	fmf0 getSetPixelHeight;
	fmf5u setThinLineAA;
	fmf5u setThinEllipseAA;
	fmf3uf blendRGBA;


	fmf3 triggerThread;
	fmfiIi waitForThread;
	fmf1 isThreadActive;
	fmf1 getThreadRetVal;
	fmf1 terminateThread;

    fmf1 createSync;
    fmf2I waitForSync;
    fmf0 deleteSync;


    fmf0 createCriticalSection;
#ifdef HAVE_XP
    fmf1 createCriticalSectionAndSpinCount;
    fmf2 setCriticalSectionSpinCount;
#endif
#ifdef HAVE_VISTA
    fmf2 createCriticalSectionEx;
#endif
    fmf1n enterCriticalSection;
#if _WIN32_WINNT >= 0x0400 
    fmf1 tryEnterCriticalSection;
#endif
    fmf1n leaveCriticalSection;
    fmf1n deleteCriticalSection;

    fmf1 interlockedIncrement;
    fmf1 interlockedDecrement;
    fmf2 interlockedExchange;
    fmf2 interlockedExchangeAdd;
    fmf3 interlockedCompareExchange;

    fmf3 triggerEvent;
    fmf5 requestRect;
    fmf0 restoreRect;

    fmf1 loadLib;
    fmf1 freeLib;
	fmf2 getLibFn;
    fmf1v callLibFmc;
    fmf1vd fcallLibFmc;
	fmf1v callLib;
	fmf1vd fcallLib;

    fmf1 getCpuReg;
    fmf0n finit;
    fmf0 fstsw;
    fmf0 fstcw;
    fmf1n fldcw;  

	fmf1 checkArrayAlloc;

	fmf1 getArrayAddress;
	fmf1 getBufferAddress;

    fmf1 setPreviewDragMode;
    fmf0 doEvents;

    fmf5f hermiteInterpolate;
    fmf5f cubicInterpolate;
    fmf3f cosineInterpolate;
    fmf3f linearInterpolate;

    fmf1 getImageTitle;



    wf0h createPopupMenu; //harry has fmf0 ????
    wf0h createMenu; //ditto???
    fmf1 destroyMenu;
    fmf5 insertMenuItem;
    fmf5 trackPopupMenu;
    fmf1 setMenu;

    fmf3 setCtlEditSize;
    fmf1 setCtlFocus;
    fmf1 setEditBoxUpdate;
    fmf1 checkCtlFocus;
	fmf0 checkDialogFocus;

    fmf3 findFirstFile;
    fmf3 findNextFile;
    fmf1 findClose;

    fmf1u isqrt;
    fmf0 isqrt_fast;	

    fmf1 getCtlText;
	fmf1 getCtlClass;
	fmf1 getCtlItemCount;
	fmf1 getCtlItemTop;
	fmf2 setCtlItemTop;
	fmf3 getCtlItemText;
	fmf3 setCtlItemText;
	fmf2 deleteCtlItem;
	fmf1 deleteCtlItems;
	fmf3 fillDir;

	fmf0 getAppTheme;
	fmf1 setDialogTheme;
	fmf2 setCtlTheme;

	fmf1 setDialogEvent;
	fmf1 clearDialogEvent;

	fmf2 getSpecialFolder;
	
	fmf2 setCtlStyle;
	fmf2 setCtlStyleEx;
	fmf2 clearCtlStyle;
	fmf2 clearCtlStyleEx;

	fmf3 setCtlBuddyStyle;
	fmf3 setCtlBuddyStyleEx;
	fmf3 clearCtlBuddyStyle;
	fmf3 clearCtlBuddyStyleEx;

	fmf1 getSysColor;
	fmf2 setCtlSysColor;
    fmf2 setCtlFontSysColor;
	fmf1 setDialogSysColor;

	fmf1 getSysMem;

	fmf2 setToolTipDelay;

	fmf2dd fr2x;
	fmf2dd fr2y;
	fmf2dd fc2m;
	fmf2dd fc2d;
	fmf1 setEndCaps;
	fmf1 setJoin;
	fmf1 setInsideFrame;
	fmf6iud setAngleArcFill;
	fmf6iud setAngleArc;

	fmf2 getDialogPos;

	fmf3 setCtlTab;
    fmf2 getCtlTab;

	fmf2 setCtlAnchor;
    fmf2 updateAnchors;
    fmf1 lockCtlScaling;
    fmf2 scaleCtls;

    fmf1 allocHost;
    fmf1 lockHost;
    fmf1 freeHost;

    fmf2 setCtlScripting;
    fmf1 checkScriptVal;
    fmf2 enableScriptVal;
    fmf1 getScriptVal;
    fmf2 setScriptVal;

    fmf0 getDialogHandle;
    fmf1 getCtlHandle;


    //fmf1u isqrt;
    fmf7 cnvX;
    fmf7 cnvY;
    fmf3 pget;
    fmf3 tget;
    fmf3 t2get;
	fmf3 t3get;
	fmf3 t4get;
    fmf4 pset;
    fmf4 tset;
    fmf4 t2set;
	fmf4 t3set;
	fmf4 t4set;
    fmf3 pgetr;
    fmf3 tgetr;
    fmf3 t2getr;
	fmf3 t3getr;
	fmf3 t4getr;
    fmf4 psetr;
    fmf4 tsetr;
    fmf4 t2setr;
	fmf4 t3setr;
	fmf4 t4setr;
    fmf0 doForEveryRow;
    fmf1 doAction;
    fmf0 abort;
    fmf0 testAbort;
    fmf2 updateProgress;
    fmf1 updatePreview;
    fmf0 getFilterInstallDir;
    fmf1 appendEllipsis;
    fmf1 stripEllipsis;
    fmf1 formatString;
    fmf2 setCtlVal;
    fmf2 setCtlDefVal;
    fmf2 setCtlText;
    fmf2v setCtlTextv;
    fmf1 enableToolTipBalloon;
    fmf3 setCtlToolTip;
    fmf3 setCtlImage;
	fmf1 getCtlImage;
    fmf2 setCtlTicFreq;
    fmf2 setCtlFontColor;
    fmf2 setCtlColor;
    fmf1 getCtlColor;
    fmf2 setCtlProperties;
    fmf2 clearCtlProperties;
    fmf3 setCtlRange;
    fmf2 setCtlLineSize;
    fmf2 setCtlPageSize;

	fmf2 setCtlThumbSize;

    fmf2 setCtlOrder;
    fmf5 createFont;
    fmf1 deleteFont;
    fmf2 setCtlFont;

    fmf5 setCtlPos;
    fmf5 setCtlPixelPos;
    fmf2 setCtlAction;
    fmf2 enableCtl;
    fmf11 createCtl;
    fmf1 initCtl;
    fmf1 deleteCtl;
    fmf0 resetAllCtls;
    fmf5 setDialogPos;
    fmf1 setDialogStyle;
    fmf1 clearDialogStyle;
    fmf1 setDialogStyleEx;
    fmf1 clearDialogStyleEx;
    fmf1 setDialogText;
    fmf1v setDialogTextv;
    fmf1 setDialogColor;
    fmf3 setDialogGradient;
    fmf1 setDialogImage;
    fmf2 setDialogImageMode;
    fmf1 setDialogDragMode;
    fmf1 setDialogRegion;
    fmf4 createRectRgn;
    fmf6 createRoundRectRgn;
    fmf3 createCircularRgn;
    fmf4 createEllipticRgn;
    fmf8v createPolyRgn;
    fmf0 createDialogRgn;
    fmf0 createTitlebarRgn;
    fmf0 createMenubarRgn;
    fmf0 createClientRgn;
    fmf0 createStatusbarRgn;
    fmf1 playSoundWave;
    fmf1 playSoundWaveLoop;
    fmf1 playSoundWaveSync;
    fmf3v msgBox;
    fmf1v Info;
    fmf1v Warn;
    fmf1v Error;
    fmf1v ErrorOk;
    fmf1v YesNo;
    fmf1v YesNoCancel;
    fmf1 sleep;
    fmf3 RGB;
    fmf4 RGBA;
    fmf1 Rval;
    fmf1 Gval;
    fmf1 Bval;
    fmf1 Aval;
    fmf15 getOpenFileName;
    fmf15 getSaveFileName;
    fmf2v chooseColor;
    fmf4 shellExec;
    fmf1 HDBUsToPixels;
    fmf1 VDBUsToPixels;
    fmfdi setGamma;
    fmf1 gamma;
    fmf1  setRegRoot;
    fmf1  getRegRoot;
    fmf1v setRegPath;
    fmf2  getRegPath;
    fmf2v putRegInt;
    fmf2v getRegInt;
    fmf2v putRegString;
    fmf3v getRegString;
    fmf3v putRegData;
    fmf3v getRegData;
    fmf3  enumRegSubKey;
    fmf5  enumRegValue;
    fmf1v deleteRegSubKey;
    fmf1v deleteRegValue;
    /* Vectors to public C-RTL functions */
    /* from <locale.h> */
    fmf4 getLocaleInfo;
    wf0U getSystemDefaultLCID;  //harry has fmf0???
    wf0U getUserDefaultLCID;  //harry has fmf0???
    /* from <time.h> */
    fmf0 clock;
    fmf1 strdate;
    fmf1 strtime;
    fmf1 time;
    /* from <string.h> */
    fmf2 strcat;
    fmf2 strchr;
    fmf2 strcmp;
    fmf2 strcoll;
    fmf2 strcpy;
    fmf2 strcspn;
    fmf1 strdup;    //needs free()
    fmf1 strerror;
    fmf1 _strerror; //not the same as strerror!
    fmf2 stricmp;
    fmf2 stricoll;
    fmf1 strlen;
    fmf1 strlwr;
    fmf3 strncat;
    fmf3 strncmp;
    fmf3 strncoll;
    fmf3 strncpy;
    fmf3 strnicmp;
    fmf3 strnicoll;
    fmf3 strnset;
    fmf2 strpbrk;
    fmf2 strrchr;
    fmf1 strrev;
    fmf2 strset;
    fmf2 strspn;
    fmf2 strstr;
    fmf2 strtok;
    fmf1 strupr;
    fmf3 strxfrm;
    /* from <memory.h> */
    fmf4 memccpy;
    fmf3 memchr;
    fmf3 memcmp;
    fmf3 memcpy;
    fmf3 memicmp;
    fmf3 memmove;
    fmf3 memset;
    /* from <math.h> */
    fmf1dd math_fabs;
    fmf1dd math_ceil;
    fmf1dd math_floor;
    fmf1dd math_round;
    fmf1dd math_chop;
    fmf2dd math_fmod;
    fmf1dd math_exp;
    fmf1dd math_log;
    fmf1dd math_log10;
    fmfdid math_ldexp;
    fmf2dd math_pow;
    fmf1dd math_sqrt;
    fmf1dd math_cos;  //fcos for now
    fmf1dd math_sin;  //fsin for now
    fmf1dd math_tan;  //ftan for now
    fmf1dd math_acos;
    fmf1dd math_asin;
    fmf1dd math_atan;
    fmf2dd math_atan2;
    fmf1dd math_cosh;
    fmf1dd math_sinh;
    fmf1dd math_tanh;
    fmf2dd math_hypot;    //non-ANSI
    /* from <malloc.h> */
  //fmf1 alloca;
    fmf2 calloc;
    fmf2 expand;
    fmf1 free;
    fmf1 malloc;
    fmf1 msize;
    fmf2 realloc;
    /* from <stdlib.h> */
    fmf0 stdlib_rand;
    fmf1un stdlib_srand;
    fmf2d strtod;
    fmf3 strtol;
    fmf3 strtoul;
    /* from <stdio.h> */
    fmf1  stdio_clearerr;
    fmf1  stdio_fclose;
    fmf0  stdio_fcloseall;   //non-ANSI
    fmf1  stdio_feof;
    fmf1  stdio_ferror;
    fmf1  stdio_fflush;
    fmf1  stdio_fgetc;
    fmf2  stdio_fgetpos;
    fmf3  stdio_fgets;
    fmf0  stdio_flushall;    //non-ANSI
    fmf2  stdio_fopen;
    fmf2v stdio_fprintf;
    fmf2  stdio_fputc;
    fmf2  stdio_fputs;
    fmf4  stdio_fread;
    fmf3  stdio_freopen;
    fmf2v stdio_fscanf;
    fmf3  stdio_fseek;
    fmf2  stdio_fsetpos;
    fmf1  stdio_ftell;
    fmf4  stdio_fwrite;
    fmf1  stdio_getc;
  //fmf0  stdio_getchar;
  //fmf1  stdio_gets;
  //fmf1  stdio_perror;
    fmf1v stdio_printf;
    fmf2  stdio_putc;
  //fmf1  stdio_putchar;
  //fmf1  stdio_puts;
    fmf1  stdio_remove;
    fmf2  stdio_rename;
    fmf1  stdio_rewind;
  //fmf1v stdio_scanf;
    fmf3v stdio_snprintf;
    fmf2v stdio_sprintf;
    fmf2v stdio_sscanf;
    fmf0  stdio_tmpfile;
    fmf1  stdio_tmpnam;
    fmf2  stdio_ungetc;

	fmf1 direct_chdir;
	fmf2 direct_getcwd;
	fmf1 direct_mkdir;
	fmf1 direct_rmdir;

    /* from <errno.h> */
    fmf0  fm_errno;     //actually int*()(void)	//X64??

    /* Vectors to private C-RTL functions */
    fmf0 _ftol;
	fmf0 _ftol2;
	fmf0 _ftol2_sse;
	fmf0 fm_chkstk;
    /* Vectors to public Win32 API functions */
    wf4 CombineRgn;
    /* Vectors to private FF+ built-in functions */
    fmf0 divCheck;
    /* Vectors to new FF+/FML built-in math functions */
    fmfdi math_iceil;
    fmfdi math_ifloor;
    fmfdi math_iround;
    fmfdi math_ichop;
    fmf2 math_ipow;
    fmfdid math_powi;
    /* Vectors to event handlers */
    fmf0 onInitDialog;
    fmf0 onCtl;
    fmf0 onZeroDivide;
    fmf0 onDivideOverflow;
    fmf0 onFilterStart;
    fmf0 onFilterEnd;
    fmf0 UserFunctions;     //really needed???
    /* Anonymous storage cells for put() and get() */       /*GLOBAL?*/
    int cell[N_CELLS];			//X64??
    /* Cached settings of user controls */                  /*GLOBAL!*/
    int pre_ctl[N_CTLS];
#ifndef TRIG_ON_FLY
    // Precomputed trig tables                              /*GLOBAL?*/
    int pre_sin[1024];				//X64??
    int pre_tan[512];
#endif
    // Precomputed gamma table
    int pre_gamma[256];                     /*TLS!! -- will vary */	//X64??
}
#if TARGET == 64
FMcontext64;
#else //TARGET == 32
FMcontext32;
#endif

    /* Global variables. */				//Should go elsewhere????

#if TARGET == 64 && defined(_WIN64) || TARGET == 32 && !defined(_WIN64)
// not cross-compiling...
FMcontext * const gFmcp;
#endif
